/*
	Copyright (C) 2004 - 2007 rt2x00 SourceForge Project
	<http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
	Module: rt2x00lib
	Abstract: rt2x00 generic device routines.
	Supported chipsets: RT2460, RT2560, RT2570,
	rt2561, rt2561s, rt2661, rt2571W & rt2671.
 */

/*
 * Set enviroment defines for rt2x00.h
 */
#define DRV_NAME "rt2x00lib"

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/etherdevice.h>

#include "rt2x00.h"

/*
 * Radio control handlers.
 */
int rt2x00lib_enable_radio(struct rt2x00_dev *rt2x00dev)
{
	int status;

	/*
	 * Don't enable the radio twice.
	 * or if the hardware button has been disabled.
	 */
	if (GET_FLAG(rt2x00dev, DEVICE_ENABLED_RADIO))
		return 0;

	status = rt2x00dev->ops->lib->enable_radio(rt2x00dev);
	if (status)
		return status;

	SET_FLAG(rt2x00dev, DEVICE_ENABLED_RADIO);

	rt2x00dev->ops->lib->toggle_rx(rt2x00dev, 1);

	ieee80211_start_queues(rt2x00dev->hw);

	return 0;
}
EXPORT_SYMBOL_GPL(rt2x00lib_enable_radio);

void rt2x00lib_disable_radio(struct rt2x00_dev *rt2x00dev)
{
	if (!GET_FLAG(rt2x00dev, DEVICE_ENABLED_RADIO))
		return;

	ieee80211_stop_queues(rt2x00dev->hw);

	rt2x00dev->ops->lib->toggle_rx(rt2x00dev, 0);

	CLEAR_FLAG(rt2x00dev, DEVICE_ENABLED_RADIO);

	rt2x00dev->ops->lib->disable_radio(rt2x00dev);
}
EXPORT_SYMBOL_GPL(rt2x00lib_disable_radio);

/*
 * Driver initialization handlers.
 */
#define create_channel_a(__entry, __channel, __tx_power, __value)	\
({									\
	(__entry)->chan = (__channel);					\
	(__entry)->freq = ( 5000 + (5 * (__channel)));			\
	(__entry)->val = __value;					\
	(__entry)->flag =						\
		IEEE80211_CHAN_W_IBSS |					\
		IEEE80211_CHAN_W_ACTIVE_SCAN |				\
		IEEE80211_CHAN_W_SCAN;					\
	(__entry)->power_level = __tx_power;				\
	(__entry)->antenna_max = 0xff;					\
})

#define create_channel_bg(__entry, __channel, __tx_power, __value)	\
({									\
	(__entry)->chan = (__channel);					\
	(__entry)->freq = 2407 + (5 * (__channel));			\
	(__entry)->val = __value;					\
	(__entry)->flag =						\
		IEEE80211_CHAN_W_IBSS |					\
		IEEE80211_CHAN_W_ACTIVE_SCAN |				\
		IEEE80211_CHAN_W_SCAN;					\
	(__entry)->power_level = __tx_power;				\
	(__entry)->antenna_max = 0xff;					\
})

#define create_rate(__entry, __rate, __mask, __plcp, __flags)		\
({									\
	(__entry)->rate = (__rate);					\
	(__entry)->val =						\
		DEVICE_SET_RATE_FIELD((__rate), RATE) |			\
		DEVICE_SET_RATE_FIELD((__mask), RATEMASK) |		\
		DEVICE_SET_RATE_FIELD((__plcp), PLCP);			\
	(__entry)->flags = (__flags);					\
	(__entry)->val2 = (__entry)->val;				\
	if ((__entry)->flags & IEEE80211_RATE_PREAMBLE2)		\
		(__entry)->val2 |= DEVICE_SET_RATE_FIELD(1, PREAMBLE);	\
	(__entry)->min_rssi_ack = 0;					\
	(__entry)->min_rssi_ack_delta = 0;				\
})

static int rt2x00lib_init_hw_modes(struct rt2x00_dev *rt2x00dev,
	struct hw_mode_spec *spec)
{
	struct ieee80211_hw *hw = rt2x00dev->hw;
	struct ieee80211_hw_mode *hwmodes;
	struct ieee80211_channel *channels;
	struct ieee80211_rate *rates;
	unsigned int i;
	unsigned char tx_power;

	hwmodes = kzalloc(sizeof(*hwmodes) * spec->num_modes, GFP_KERNEL);
	if (!hwmodes)
		goto exit;

	channels = kzalloc(sizeof(*channels) * spec->num_channels, GFP_KERNEL);
	if (!channels)
		goto exit_free_modes;

	rates = kzalloc(sizeof(*rates) * spec->num_rates, GFP_KERNEL);
	if (!rates)
		goto exit_free_channels;

	/*
	 * Initialize Rate list.
	 */
	create_rate(&rates[0], 10, 0x001, 0x00, IEEE80211_RATE_CCK);
	create_rate(&rates[1], 20, 0x003, 0x01, IEEE80211_RATE_CCK_2);
	create_rate(&rates[2], 55, 0x007, 0x02, IEEE80211_RATE_CCK_2);
	create_rate(&rates[3], 110, 0x00f, 0x03, IEEE80211_RATE_CCK_2);

	if (spec->num_rates > 4) {
		create_rate(&rates[4], 60, 0x01f, 0x0b, IEEE80211_RATE_OFDM);
		create_rate(&rates[5], 90, 0x03f, 0x0f, IEEE80211_RATE_OFDM);
		create_rate(&rates[6], 120, 0x07f, 0x0a, IEEE80211_RATE_OFDM);
		create_rate(&rates[7], 180, 0x0ff, 0x0e, IEEE80211_RATE_OFDM);
		create_rate(&rates[8], 240, 0x1ff, 0x09, IEEE80211_RATE_OFDM);
		create_rate(&rates[9], 360, 0x3ff, 0x0d, IEEE80211_RATE_OFDM);
		create_rate(&rates[10], 480, 0x7ff, 0x08, IEEE80211_RATE_OFDM);
		create_rate(&rates[11], 540, 0xfff, 0x0c, IEEE80211_RATE_OFDM);
	}

	/*
	 * Initialize Channel list.
	 */
	for (i = 0; i < 14; i++)
		create_channel_bg(&channels[i], i + 1,
			spec->tx_power_bg[i], spec->chan_val_bg[i]);

	if (spec->num_channels > 14) {
		for (i = 14; i < spec->num_channels; i++) {
			if (i < 22)
				channels[i].chan = 36;
			else if (i < 33)
				channels[i].chan = 100;
			else
				channels[i].chan = 149;
			channels[i].chan += ((i - 14) * 4);

			if (spec->tx_power_a)
				tx_power = spec->tx_power_a[i];
			else
				tx_power = spec->tx_power_default;

			create_channel_a(&channels[i], channels[i].chan,
				tx_power, spec->chan_val_a[i]);
		}
	}

	/*
	 * Intitialize 802.11b
	 * Rates: CCK.
	 * Channels: OFDM.
	 */
	if (spec->num_modes > HWMODE_B) {
		hwmodes[HWMODE_B].mode = MODE_IEEE80211B;
		hwmodes[HWMODE_B].num_channels = 14;
		hwmodes[HWMODE_B].num_rates = 4;
		hwmodes[HWMODE_B].channels = channels;
		hwmodes[HWMODE_B].rates = rates;
	}

	/*
	 * Intitialize 802.11g
	 * Rates: CCK, OFDM.
	 * Channels: OFDM.
	 */
	if (spec->num_modes > HWMODE_G) {
		hwmodes[HWMODE_G].mode = MODE_IEEE80211G;
		hwmodes[HWMODE_G].num_channels = 14;
		hwmodes[HWMODE_G].num_rates = spec->num_rates;
		hwmodes[HWMODE_G].channels = channels;
		hwmodes[HWMODE_G].rates = rates;
	}

	/*
	 * Intitialize 802.11a
	 * Rates: OFDM.
	 * Channels: OFDM, UNII, HiperLAN2.
	 */
	if (spec->num_modes > HWMODE_A) {
		hwmodes[HWMODE_A].mode = MODE_IEEE80211A;
		hwmodes[HWMODE_A].num_channels = spec->num_channels - 14;
		hwmodes[HWMODE_A].num_rates = spec->num_rates - 4;
		hwmodes[HWMODE_A].channels = &channels[14];
		hwmodes[HWMODE_A].rates = &rates[4];
	}

	if (spec->num_modes > HWMODE_G &&
	    ieee80211_register_hwmode(hw, &hwmodes[HWMODE_G]))
		goto exit_free_rates;

	if (spec->num_modes > HWMODE_B &&
	    ieee80211_register_hwmode(hw, &hwmodes[HWMODE_B]))
		goto exit_free_rates;

	if (spec->num_modes > HWMODE_A &&
	    ieee80211_register_hwmode(hw, &hwmodes[HWMODE_A]))
		goto exit_free_rates;

	rt2x00dev->hwmodes = hwmodes;

	return 0;

exit_free_rates:
	kfree(rates);

exit_free_channels:
	kfree(channels);

exit_free_modes:
	kfree(hwmodes);

exit:
	ERROR("Allocation ieee80211 modes failed.\n");
	return -ENOMEM;
}

static void rt2x00lib_deinit_hw(struct rt2x00_dev *rt2x00dev)
{
	if (GET_FLAG(rt2x00dev, DEVICE_INITIALIZED_HW))
		ieee80211_unregister_hw(rt2x00dev->hw);

	if (likely(rt2x00dev->hwmodes)) {
		kfree(rt2x00dev->hwmodes->channels);
		kfree(rt2x00dev->hwmodes->rates);
		kfree(rt2x00dev->hwmodes);
		rt2x00dev->hwmodes = NULL;
	}
}

static int rt2x00lib_init_hw(struct rt2x00_dev *rt2x00dev)
{
	struct hw_mode_spec *spec = &rt2x00dev->spec;
	int status;

	/*
	 * Initialize device.
	 */
	SET_IEEE80211_DEV(rt2x00dev->hw, rt2x00dev->device);

	/*
	 * Initialize MAC address.
	 */
	if (!is_valid_ether_addr(spec->mac_addr)) {
		ERROR("Invalid MAC addr: " MAC_FMT ".\n", MAC_ARG(spec->mac_addr));
		return -EINVAL;
	}
	rt2x00dev->ops->lib->config_mac_addr(rt2x00dev, spec->mac_addr);
	SET_IEEE80211_PERM_ADDR(rt2x00dev->hw, spec->mac_addr);

	/*
	 * Initialize HW modes.
	 */
	status = rt2x00lib_init_hw_modes(rt2x00dev, spec);
	if (status)
		return status;

	/*
	 * Register HW.
	 */
	status = ieee80211_register_hw(rt2x00dev->hw);
	if (status) {
		rt2x00lib_deinit_hw(rt2x00dev);
		return status;
	}

	SET_FLAG(rt2x00dev, DEVICE_INITIALIZED_HW);

	return 0;
}

#ifdef CONFIG_RT2X00_DEBUGFS
static void rt2x00lib_open_debugfs(struct rt2x00_dev *rt2x00dev)
{
	rt2x00dev->debugfs_intf = rt2x00debug_register(
		rt2x00dev->ops->debugfs, rt2x00dev->hw->wiphy, rt2x00dev);
	if (!rt2x00dev->debugfs_intf)
		ERROR("Failed to register debug handler.\n");
}

static void rt2x00lib_close_debugfs(struct rt2x00_dev *rt2x00dev)
{
	rt2x00debug_deregister(rt2x00dev->debugfs_intf);
}
#else /* CONFIG_RT2X00_DEBUGFS */
static inline void rt2x00lib_open_debugfs(struct rt2x00_dev *rt2x00dev){}
static inline void rt2x00lib_close_debugfs(struct rt2x00_dev *rt2x00dev){}
#endif /* CONFIG_RT2X00_DEBUGFS */

#ifdef CONFIG_RT2X00_LIB_FIRMWARE
struct fw_entry {
	u32	chip;
	char	*name;
};

static void rt2x00lib_load_firmware_continued(const struct firmware *fw,
	void *context)
{
	struct rt2x00_dev *rt2x00dev = context;

	if (!fw || !fw->size || !fw->data) {
		ERROR("Failed to read Firmware.\n");
		SET_FLAG(rt2x00dev, FIRMWARE_FAILED);
		return;
	}

	if (rt2x00dev->ops->lib->load_firmware(rt2x00dev, fw)) {
		SET_FLAG(rt2x00dev, FIRMWARE_FAILED);
		return;
	}

	/*
	 * Initialize ieee80211 structure.
	 */
	if (rt2x00lib_init_hw(rt2x00dev)) {
		ERROR("Failed to initialize hw.\n");
		rt2x00lib_free_dev(rt2x00dev);
		return;
	}

	/*
	 * Open the debugfs entry.
	 */
	rt2x00lib_open_debugfs(rt2x00dev);

	SET_FLAG(rt2x00dev, FIRMWARE_LOADED);
}

static int rt2x00lib_load_firmware(struct rt2x00_dev *rt2x00dev)
{
	unsigned int i;
	int status = -EINVAL;
	static const struct fw_entry fw_list[] = {
		{ RT2561,	"rt2561.bin" },
		{ RT2561s,	"rt2561s.bin" },
		{ RT2661,	"rt2661.bin" },
		{ RT2571,	"rt73.bin" },
	};

	/*
	 * Read correct firmware from harddisk.
	 */
	for (i = 0; i < ARRAY_SIZE(fw_list); i++) {
		if (!rt2x00_rt(&rt2x00dev->chip, fw_list[i].chip))
			continue;
		status = request_firmware_nowait(THIS_MODULE,
			FW_ACTION_HOTPLUG, fw_list[i].name, rt2x00dev->device,
			rt2x00dev, &rt2x00lib_load_firmware_continued);
	}

	if (status)
		ERROR("Failed to request Firmware.\n");

	return status;

}

int rt2x00lib_load_firmware_wait(struct rt2x00_dev *rt2x00dev)
{
	unsigned int i;

	for (i = 0; i < 150; i++) {
		if (GET_FLAG(rt2x00dev, FIRMWARE_FAILED))
			return -EIO;
		if (GET_FLAG(rt2x00dev, FIRMWARE_LOADED))
			return 0;
		msleep(20);
	}

	ERROR("Firmware loading timed out.\n");
	return -ETIMEDOUT;
}
EXPORT_SYMBOL_GPL(rt2x00lib_load_firmware_wait);
#else /* CONFIG_RT2X00_LIB_FIRMWARE */
int rt2x00lib_load_firmware(struct rt2x00_dev *rt2x00dev) 
{
	ERROR("Firmware loading requested by non-firmware driver.\n");
	return -EINVAL;
}
#endif /* CONFIG_RT2X00_LIB_FIRMWARE */

/*
 * driver allocation handlers.
 */
static int rt2x00lib_alloc_rings(struct rt2x00_dev *rt2x00dev)
{
	unsigned int ring_num;
	unsigned int i;

	/*
	 * We support 1 RX queue, 1 Beacon ring
	 * and we could support 1 ATIM ring
	 * if the driver has raised that flag.
	 */
	ring_num = 2 + rt2x00dev->hw->queues +
		GET_FLAG(rt2x00dev, DEVICE_SUPPORT_ATIM);

	rt2x00dev->ring = kzalloc(
		sizeof(struct data_ring) * ring_num, GFP_KERNEL);
	if (!rt2x00dev->ring) {
		ERROR("Ring allocation failed.\n");
		return -ENOMEM;
	}

	for (i = 0; i < ring_num; i++) {
		rt2x00dev->ring[i].rt2x00dev = rt2x00dev;

		/*
		 * Initialize ring parameters.
		 * cw_min: 2^5 = 32.
		 * cw_max: 2^10 = 1024.
		 */
		rt2x00dev->ring[i].tx_params.aifs = 2;
		rt2x00dev->ring[i].tx_params.cw_min = 5;
		rt2x00dev->ring[i].tx_params.cw_max = 10;
	}

	return 0;
}

int rt2x00lib_alloc_dev(struct rt2x00_dev *rt2x00dev)
{
	/*
	 * Create workqueue.
	 */
	rt2x00dev->workqueue = create_singlethread_workqueue(DRV_NAME);
	if (!rt2x00dev->workqueue)
		goto exit;

	/*
	 * Initialize configuration work.
	 */
	INIT_DELAYED_WORK(&rt2x00dev->link.work,
		rt2x00dev->ops->lib->link_tuner);

	/*
	 * Reset current working type.
	 */
	rt2x00dev->interface.type = -EINVAL;

	/*
	 * Allocate ring array.
	 */
	if (rt2x00lib_alloc_rings(rt2x00dev))
		goto exit;

	/*
	 * Check if we need to load the firmware.
	 */
	if (rt2x00dev->ops->lib->load_firmware) {
		/*
		 * Request firmware and wait with further
		 * initializing of the card until the firmware
		 * has been loaded.
		 */
		if (rt2x00lib_load_firmware(rt2x00dev))
			goto exit;
	} else {
		/*
		 * No firmware needed, just set the flag
		 * as if it has been set correctly.
		 */
		SET_FLAG(rt2x00dev, FIRMWARE_LOADED);

		/*
		 * Initialize ieee80211 structure.
		 */
		if (rt2x00lib_init_hw(rt2x00dev)) {
			ERROR("Failed to initialize hw.\n");
			goto exit;
		}

		/*
		 * Open the debugfs entry.
		 */
		rt2x00lib_open_debugfs(rt2x00dev);
	}

	return 0;

exit:
	rt2x00lib_free_dev(rt2x00dev);

	return -ENODEV;
}
EXPORT_SYMBOL_GPL(rt2x00lib_alloc_dev);

void rt2x00lib_free_dev(struct rt2x00_dev *rt2x00dev)
{
	/*
	 * Close debugfs entry.
	 */
	rt2x00lib_close_debugfs(rt2x00dev);

	/*
	 * Free workqueue.
	 */
	if (likely(rt2x00dev->workqueue)) {
		destroy_workqueue(rt2x00dev->workqueue);
		rt2x00dev->workqueue = NULL;
	}

	/*
	 * Free ring structures.
	 */
	kfree(rt2x00dev->ring);
	rt2x00dev->ring = NULL;

	/*
	 * Free EEPROM memory.
	 */
	kfree(rt2x00dev->eeprom);
	rt2x00dev->eeprom = NULL;

	/*
	 * Free ieee80211_hw memory.
	 */
	rt2x00lib_deinit_hw(rt2x00dev);
}
EXPORT_SYMBOL_GPL(rt2x00lib_free_dev);

/*
 * Interrupt context handlers.
 */
void rt2x00lib_txdone(struct data_entry *entry,
	const int status, const int retry)
{
	struct rt2x00_dev *rt2x00dev = entry->ring->rt2x00dev;
	struct ieee80211_tx_status *tx_status = &entry->tx_status;
	struct ieee80211_low_level_stats *stats = &rt2x00dev->low_level_stats;

	/*
	 * Update TX statistics.
	 */
	tx_status->flags = 0;
	tx_status->ack_signal = 0;
	tx_status->excessive_retries = (status == TX_FAIL_RETRY);
	tx_status->retry_count = retry;

	if (!(tx_status->control.flags & IEEE80211_TXCTL_NO_ACK)) {
		if (status == TX_SUCCESS || status == TX_SUCCESS_RETRY)
			tx_status->flags |= IEEE80211_TX_STATUS_ACK;
		else
			stats->dot11ACKFailureCount++;
	}

	tx_status->queue_length = entry->ring->stats.limit;
	tx_status->queue_number = tx_status->control.queue;

	if (tx_status->control.flags & IEEE80211_TXCTL_USE_RTS_CTS) {
		if (status == TX_SUCCESS || status == TX_SUCCESS_RETRY)
			stats->dot11RTSSuccessCount++;
		else
			stats->dot11RTSFailureCount++;
	}

	/*
	 * Send the tx_status to mac80211,
	 * that method also cleans up the skb structure.
	 */
	ieee80211_tx_status_irqsafe(rt2x00dev->hw, entry->skb, tx_status);

	entry->skb = NULL;
}
EXPORT_SYMBOL_GPL(rt2x00lib_txdone);

void rt2x00lib_rxdone(struct data_entry *entry, char *data,
	const int size, const int signal, const int rssi, const int ofdm)
{
	struct rt2x00_dev *rt2x00dev = entry->ring->rt2x00dev;
	struct ieee80211_rx_status *rx_status = &rt2x00dev->rx_status;
	struct ieee80211_hw_mode *mode;
	struct ieee80211_rate *rate;
	struct sk_buff *skb;
	unsigned int i;
	int val = 0;

	/*
	 * Update RX statistics.
	 */
	mode = &rt2x00dev->hwmodes[rt2x00dev->curr_hwmode];
	for (i = 0; i < mode->num_rates; i++) {
		rate = &mode->rates[i];

		/*
		 * When frame was received with an OFDM bitrate,
		 * the signal is the PLCP value. If it was received with
		 * a CCK bitrate the signal is the rate in 0.5kbit/s.
		 */
		if (!ofdm)
			val = DEVICE_GET_RATE_FIELD(rate->val, RATE);
		else
			val = DEVICE_GET_RATE_FIELD(rate->val, PLCP);

		if (val == signal) {
			/*
			 * Check for preamble bit.
			 */
			if (signal & 0x08)
				val = rate->val2;
			val = rate->val;
			break;
		}
	}

	rx_status->rate = val;
	rx_status->ssi = rssi;
	rx_status->noise = rt2x00_get_link_noise(&rt2x00dev->link);
	rt2x00_update_link_rssi(&rt2x00dev->link, rssi);

	/*
	 * Let's allocate a sk_buff where we can store the received data in,
	 * note that if data is NULL, we still have to allocate a sk_buff
	 * but that we should use that to replace the sk_buff which is already
	 * inside the entry.
	 */
	skb = dev_alloc_skb(size + NET_IP_ALIGN);
	if (!skb)
		return;

	skb_reserve(skb, NET_IP_ALIGN);
	skb_put(skb, size);

	if (data) {
		memcpy(skb->data, data, size);
		entry->skb = skb;
		skb = NULL;
	}

	ieee80211_rx_irqsafe(rt2x00dev->hw, entry->skb, rx_status);
	entry->skb = skb;
}
EXPORT_SYMBOL_GPL(rt2x00lib_rxdone);

/*
 * TX descriptor initializer
 */
void rt2x00lib_write_tx_desc(struct rt2x00_dev *rt2x00dev,
	struct data_entry *entry, struct data_desc *txd,
	struct ieee80211_hdr *ieee80211hdr, unsigned int length,
	struct ieee80211_tx_control *control)
{
	struct data_entry_desc desc;
	int tx_rate;
	int bitrate;
	int duration;
	int residual;
	u16 frame_control;
	u16 seq_ctrl;

	/*
	 * Identify queue
	 */
	if (control->queue < rt2x00dev->hw->queues)
		desc.queue = control->queue;
	else
		desc.queue = 15;

	/*
	 * Read required fields from ieee80211 header.
	 */
	frame_control = le16_to_cpu(ieee80211hdr->frame_control);
	seq_ctrl = le16_to_cpu(ieee80211hdr->seq_ctrl);

	tx_rate = control->tx_rate;

	/*
	 * Check if this is a rts frame
	 */
	if (is_rts_frame(frame_control)) {
		entry->reg |= ENTRY_TXD_RTS_FRAME;
		if (control->rts_cts_rate)
			tx_rate = control->rts_cts_rate;
	}

	/*
	 * Check for OFDM
	 */
	if (DEVICE_GET_RATE_FIELD(tx_rate, RATEMASK) & DEV_OFDM_RATE)
		entry->reg |= ENTRY_TXD_OFDM_RATE;

	/*
	 * Check if more fragments are pending
	 */
	if (ieee80211_get_morefrag(ieee80211hdr))
		entry->reg |= ENTRY_TXD_MORE_FRAG;

	/*
	 * Check if this is a new sequence
	 */
	if ((seq_ctrl & IEEE80211_SCTL_FRAG) == 0)
		entry->reg |= ENTRY_TXD_NEW_SEQ;

	/*
	 * Beacons and probe responses require the tsf timestamp
	 * to be inserted into the frame.
	 */
	if (control->queue == IEEE80211_TX_QUEUE_BEACON ||
	    is_probe_resp(frame_control))
		entry->reg |= ENTRY_TXD_REQ_TIMESTAMP;

	/*
	 * Check if ACK is required
	 */
	if (!(control->flags & IEEE80211_TXCTL_NO_ACK))
		entry->reg |= ENTRY_TXD_REQ_ACK;

	/*
	 * Determine with what IFS priority this frame should be send.
	 * Set ifs to IFS_SIFS when the this is not the first fragment,
	 * or this fragment came after RTS/CTS.
	 */
	if ((seq_ctrl & IEEE80211_SCTL_FRAG) > 0 ||
	    entry->reg & ENTRY_TXD_RTS_FRAME)
		desc.ifs = IFS_SIFS;
	else
		desc.ifs = IFS_BACKOFF;

	/*
	 * How the length should be processed depends
	 * on if we are working with OFDM rates or not.
	 */
	if (entry->reg & ENTRY_TXD_OFDM_RATE) {
		residual = 0;
		desc.length_high = ((length + FCS_LEN) >> 6) & 0x3f;
		desc.length_low = ((length + FCS_LEN) & 0x3f);

	} else {
		bitrate = DEVICE_GET_RATE_FIELD(tx_rate, RATE);

		/*
		 * Convert length to microseconds.
		 */
		residual = get_duration_res(length + FCS_LEN, bitrate);
		duration = get_duration(length + FCS_LEN, bitrate);

		if (residual != 0)
			duration++;

		desc.length_high = duration >> 8;
		desc.length_low = duration & 0xff;
	}

	/*
	 * Create the signal and service values.
	 */
	desc.signal = DEVICE_GET_RATE_FIELD(tx_rate, PLCP);
	if (DEVICE_GET_RATE_FIELD(tx_rate, PREAMBLE))
		desc.signal |= 0x08;

	desc.service = 0x04;
	if (residual <= (8 % 11))
		desc.service |= 0x80;

	rt2x00dev->ops->lib->write_tx_desc(rt2x00dev, entry, txd, &desc,
		ieee80211hdr, length, control);
}
EXPORT_SYMBOL_GPL(rt2x00lib_write_tx_desc);

/*
 * rt2x00lib module information.
 */
static char version[] =
	DRV_NAME " - " DRV_VERSION " (" DRV_RELDATE ") by " DRV_PROJECT;

MODULE_AUTHOR(DRV_PROJECT);
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION("rt2x00 library");
MODULE_LICENSE("GPL");

static int __init rt2x00lib_init(void)
{
	printk(KERN_INFO "Loading module: %s.\n", version);
	return 0;
}

static void __exit rt2x00lib_exit(void)
{
	printk(KERN_INFO "Unloading module: %s.\n", version);
}

module_init(rt2x00lib_init);
module_exit(rt2x00lib_exit);
