/*
	Copyright (C) 2004 - 2007 rt2x00 SourceForge Project
	<http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
	Module: rt2x00usb
	Abstract: rt2x00 generic usb device routines.
	Supported chipsets: rt2570, rt2571W & rt2671.
 */

/*
 * Set enviroment defines for rt2x00.h
 */
#define DRV_NAME "rt2x00usb"

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/usb.h>

#include "rt2x00.h"
#include "rt2x00usb.h"

/*
 * Beacon handlers.
 */
int rt2x00usb_beacon_update(struct ieee80211_hw *hw, struct sk_buff *skb,
	struct ieee80211_tx_control *control)
{
	struct rt2x00_dev *rt2x00dev = hw->priv;
	struct usb_device *usb_dev =
		interface_to_usbdev(rt2x00dev_usb(rt2x00dev));
	struct data_ring *ring =
		rt2x00_get_ring(rt2x00dev, IEEE80211_TX_QUEUE_BEACON);
	struct data_entry *beacon;
	struct data_entry *guardian;
	int length;

	/*
	 * Just in case the ieee80211 doesn't set this,
	 * but we need this queue set for the descriptor
	 * initialization.
	 */
	control->queue = IEEE80211_TX_QUEUE_BEACON;

	/*
	 * Obtain 2 entries, one for the guardian byte,
	 * the second for the actual beacon.
	 */
	guardian = rt2x00_get_data_entry(ring);
	rt2x00_ring_index_inc(ring);
	beacon = rt2x00_get_data_entry(ring);

	/*
	 * First we create the beacon.
	 */
	skb_push(skb, ring->desc_size);
	rt2x00lib_write_tx_desc(rt2x00dev, beacon,
		(struct data_desc*)skb->data,
		(struct ieee80211_hdr*)(skb->data + ring->desc_size),
		skb->len - ring->desc_size,
		control);

	/*
	 * Length passed to usb_fill_urb cannot be an odd number,
	 * so add 1 byte to make it even.
	 */
	length = skb->len;
	if (length % 2)
		length++;

	usb_fill_bulk_urb(
		beacon->priv,
		usb_dev,
		usb_sndbulkpipe(usb_dev, 1),
		skb->data,
		length,
		rt2x00usb_beacondone,
		beacon);

	beacon->skb = skb;

	/*
	 * Second we need to create the guardian byte.
	 * We only need a single byte, so lets recycle
	 * the 'flags' field we are not using for beacons.
	 */
	guardian->reg = 0;
	usb_fill_bulk_urb(
		guardian->priv,
		usb_dev,
		usb_sndbulkpipe(usb_dev, 1),
		&guardian->reg,
		1,
		rt2x00usb_beacondone,
		guardian);

	/*
	 * Send out the guardian byte.
	 */
	usb_submit_urb(guardian->priv, GFP_ATOMIC);

	/*
	 * Enable beacon generation.
	 */
	rt2x00dev->ops->lib->kick_beacon_gen(rt2x00dev);

	return 0;
}
EXPORT_SYMBOL_GPL(rt2x00usb_beacon_update);

void rt2x00usb_beacondone(struct urb *urb)
{
	struct data_entry *entry = (struct data_entry*)urb->context;
	struct data_ring *ring = entry->ring;

	if (!GET_FLAG(ring->rt2x00dev, DEVICE_ENABLED_RADIO))
		return;

	/*
	 * Check if this was the guardian beacon,
	 * if that was the case we need to send the real beacon now.
	 * Otherwise we should free the sk_buffer, the device
	 * should be doing the rest of the work now.
	 */
	if (ring->index == 1) {
		rt2x00_ring_index_done_inc(ring);
		entry = rt2x00_get_data_entry(ring);
		usb_submit_urb(entry->priv, GFP_ATOMIC);
		rt2x00_ring_index_inc(ring);
	} else if (ring->index_done == 1) {
		entry = rt2x00_get_data_entry_done(ring);
		if (entry->skb) {
			dev_kfree_skb(entry->skb);
			entry->skb = NULL;
		}
		rt2x00_ring_index_done_inc(ring);
	}
}
EXPORT_SYMBOL_GPL(rt2x00usb_beacondone);

/*
 * TX data handlers.
 */
static void rt2x00usb_interrupt_txdone(struct urb *urb)
{
	struct data_entry *entry = (struct data_entry*)urb->context;
	struct data_ring *ring = entry->ring;
	struct rt2x00_dev *rt2x00dev = ring->rt2x00dev;
	struct data_desc *txd = (struct data_desc *)entry->skb->data;
	u32 word;
	int tx_status;

	if (!GET_FLAG(rt2x00dev, DEVICE_ENABLED_RADIO) ||
	    !GET_FLAG(entry, ENTRY_OWNER_NIC))
		return;

	CLEAR_FLAG(entry, ENTRY_OWNER_NIC);

	rt2x00_desc_read(txd, 0, &word);

	/*
	 * Remove the descriptor data from the buffer.
	 */
	skb_pull(entry->skb, ring->desc_size);

	/*
	 * Obtain the status about this packet.
	 */
	tx_status = !urb->status ? TX_SUCCESS : TX_FAIL_RETRY;

	rt2x00lib_txdone(entry, tx_status, 0);

	/*
	 * Make this entry available for reuse.
	 */
	entry->reg = 0;
	rt2x00_ring_index_done_inc(entry->ring);

	/*
	 * If the data ring was full before the txdone handler
	 * we must make sure the packet queue in the mac80211 stack
	 * is reenabled when the txdone handler has finished.
	 */
	if (!rt2x00_ring_full(ring))
		ieee80211_wake_queue(rt2x00dev->hw,
			entry->tx_status.control.queue);
}

int rt2x00usb_write_tx_data(struct rt2x00_dev *rt2x00dev,
	struct data_ring *ring, struct sk_buff *skb,
	struct ieee80211_tx_control *control)
{
	struct usb_device *usb_dev =
		interface_to_usbdev(rt2x00dev_usb(rt2x00dev));
	struct ieee80211_hdr *ieee80211hdr = (struct ieee80211_hdr*)skb->data;
	struct data_entry *entry = rt2x00_get_data_entry(ring);
	struct data_desc *txd;
	u32 length = skb->len;

	if (rt2x00_ring_full(ring)) {
		ieee80211_stop_queue(rt2x00dev->hw, control->queue);
		return -EINVAL;
	}

	if (GET_FLAG(entry, ENTRY_OWNER_NIC)) {
		ERROR("Arrived at non-free entry in the non-full queue %d.\n"
			"Please file bug report to %s.\n",
			control->queue, DRV_PROJECT);
		ieee80211_stop_queue( rt2x00dev->hw, control->queue);
		return -EINVAL;
	}

	skb_push(skb, rt2x00dev->hw->extra_tx_headroom);
	txd = (struct data_desc*)skb->data;
	rt2x00lib_write_tx_desc(rt2x00dev, entry, txd, ieee80211hdr,
		length, control);
	memcpy(&entry->tx_status.control, control, sizeof(*control));
	entry->skb = skb;

	/*
	 * Length passed to usb_fill_urb cannot be an odd number,
	 * so add 1 byte to make it even.
	 */
	length += rt2x00dev->hw->extra_tx_headroom;
	if (length % 2)
		length++;

	SET_FLAG(entry, ENTRY_OWNER_NIC);
	usb_fill_bulk_urb(
		entry->priv,
		usb_dev,
		usb_sndbulkpipe(usb_dev, 1),
		skb->data,
		length,
		rt2x00usb_interrupt_txdone,
		entry);
	usb_submit_urb(entry->priv, GFP_ATOMIC);

	rt2x00_ring_index_inc(ring);

	if (rt2x00_ring_full(ring))
		ieee80211_stop_queue(rt2x00dev->hw, control->queue);

	return 0;
}
EXPORT_SYMBOL_GPL(rt2x00usb_write_tx_data);

/*
 * Device initialization handlers.
 */
static int rt2x00usb_alloc_dma_ring(struct rt2x00_dev *rt2x00dev,
	const int queue, const u16 max_entries, const u16 data_size,
	const u16 desc_size, const int rx_ring)
{
	struct data_ring *ring = &rt2x00dev->ring[queue];
	unsigned int entry_size;
	unsigned int i;

	ring->stats.limit = max_entries;
	ring->data_size = data_size;
	ring->desc_size = desc_size;

	/*
	 * Allocate all ring entries.
	 */
	ring->entry = kzalloc(ring->stats.limit * sizeof(struct data_entry),
		GFP_KERNEL);
	if (!ring->entry)
		return -ENOMEM;

	/*
	 * Initialize all ring entries to contain valid
	 * addresses.
	 */
	entry_size = ring->data_size + ring->desc_size;
	for (i = 0; i < ring->stats.limit; i++) {
		ring->entry[i].flags = 0;
		ring->entry[i].ring = ring;
		ring->entry[i].priv = usb_alloc_urb(0, GFP_KERNEL);
		if (!ring->entry[i].priv)
			return -ENOMEM;

		if (rx_ring) {
			ring->entry[i].skb =
				dev_alloc_skb(NET_IP_ALIGN + entry_size);
			if (!ring->entry[i].skb)
				return -ENOMEM;

			skb_reserve(ring->entry[i].skb, NET_IP_ALIGN);
			skb_put(ring->entry[i].skb, entry_size);
		}
	}

	return 0;
}

static void rt2x00usb_free_ring(struct rt2x00_dev *rt2x00dev, const int queue)
{
	struct data_ring *ring = &rt2x00dev->ring[queue];
	unsigned int i;

	if (!ring->entry)
		return;

	for (i = 0; i < ring->stats.limit; i++) {
		usb_kill_urb(ring->entry[i].priv);
		usb_free_urb(ring->entry[i].priv);
		if (ring->entry[i].skb)
			kfree_skb(ring->entry[i].skb);
	}

	kfree(ring->entry);
	ring->entry = NULL;
}

static int rt2x00usb_allocate_dma_rings(struct rt2x00_dev *rt2x00dev)
{
	unsigned int i;
	int rings;

	/*
	 * The TX rings reported to mac80211,
	 * plus the additional ATIM ring (if supported).
	 */
	rings = rt2x00dev->hw->queues +
		GET_FLAG(rt2x00dev, DEVICE_SUPPORT_ATIM);

	/*
	 * First allocate the TX rings (minus BEACON).
	 */
	for (i = 0; i < rings; i++) {
		if (rt2x00usb_alloc_dma_ring(rt2x00dev, i,
			TX_ENTRIES, DATA_FRAME_SIZE, rt2x00dev->txd_size, 0))
			return -ENOMEM;
	}

	/*
	 * Allocate the BEACON ring.
	 */
	if (rt2x00usb_alloc_dma_ring(rt2x00dev, rings,
		BEACON_ENTRIES, MGMT_FRAME_SIZE, rt2x00dev->txd_size, 0))
		return -ENOMEM;

	/*
	 * Allocate the RX ring.
	 */
	if (rt2x00usb_alloc_dma_ring(rt2x00dev, rings + 1,
		RX_ENTRIES, DATA_FRAME_SIZE, rt2x00dev->rxd_size, 1))
		return -ENOMEM;

	return 0;
}

static void rt2x00usb_free_rings(struct rt2x00_dev *rt2x00dev)
{
	unsigned int i;
	int rings;

	/*
	 * The BEACON and RX ring, plus TX rings reported to mac80211,
	 * and the additional ATIM ring (if supported).
	 */
	rings = 2 + rt2x00dev->hw->queues +
		GET_FLAG(rt2x00dev, DEVICE_SUPPORT_ATIM);

	for (i = 0; i < rings; i++)
		rt2x00usb_free_ring(rt2x00dev, i);
}

int rt2x00usb_initialize(struct rt2x00_dev *rt2x00dev)
{
	if (GET_FLAG(rt2x00dev, DEVICE_INITIALIZED))
		return 0;

	/*
	 * We must wait on the firmware before
	 * we can safely continue.
	 */
	if (rt2x00lib_load_firmware_wait(rt2x00dev))
		return -ENODEV;

	/*
	 * Allocate all data rings.
	 */
	if (rt2x00usb_allocate_dma_rings(rt2x00dev)) {
		ERROR("DMA allocation failed.\n");
		goto exit_fail;
	}

	/*
	 * Reset the channel_change_time value
	 * to make sure it will be correctly initialized
	 * after the radio has been enabled.
	 */
	rt2x00dev->hw->channel_change_time = 0;

	SET_FLAG(rt2x00dev, DEVICE_INITIALIZED);

	return 0;

exit_fail:
	rt2x00usb_free_rings(rt2x00dev);

	return -EIO;
}
EXPORT_SYMBOL_GPL(rt2x00usb_initialize);

void rt2x00usb_uninitialize(struct rt2x00_dev *rt2x00dev)
{
	if (!GET_FLAG(rt2x00dev, DEVICE_INITIALIZED))
		return;

	/*
	 * Flush out all pending work.
	 */
	flush_workqueue(rt2x00dev->workqueue);

	/*
	 * Free DMA rings.
	 */
	rt2x00usb_free_rings(rt2x00dev);

	CLEAR_FLAG(rt2x00dev, DEVICE_INITIALIZED);
}
EXPORT_SYMBOL_GPL(rt2x00usb_uninitialize);

/*
 * USB driver handlers.
 */
static void rt2x00usb_remove_dev(struct rt2x00_dev *rt2x00dev)
{
	rt2x00lib_free_dev(rt2x00dev);
}

static int rt2x00usb_probe_dev(struct rt2x00_dev *rt2x00dev)
{
	int retval;

	/*
	 * Let the driver probe the device to
	 * detect the capabilities.
	 */
	retval = rt2x00dev->ops->lib->init_hw(rt2x00dev);
	if (retval) {
		ERROR("Failed to allocate device.\n");
		goto exit;
	}

	return rt2x00lib_alloc_dev(rt2x00dev);

exit:
	rt2x00usb_remove_dev(rt2x00dev);

	return retval;
}

int rt2x00usb_probe(struct usb_interface *usb_intf,
	const struct usb_device_id *id)
{
	struct usb_device *usb_dev = interface_to_usbdev(usb_intf);
	struct rt2x00_ops *ops = (struct rt2x00_ops*)id->driver_info;
	struct ieee80211_hw *hw;
	struct rt2x00_dev *rt2x00dev;
	int retval;

	usb_dev = usb_get_dev(usb_dev);

	hw = ieee80211_alloc_hw(sizeof(struct rt2x00_dev), ops->hw);
	if (!hw) {
		ERROR("Failed to allocate hardware.\n");
		retval = -ENOMEM;
		goto exit_put_device;
	}

	usb_set_intfdata(usb_intf, hw);

	rt2x00dev = hw->priv;
	rt2x00dev->dev = usb_intf;
	rt2x00dev->device = &usb_intf->dev;
	rt2x00dev->ops = ops;
	rt2x00dev->hw = hw;

	retval = rt2x00usb_probe_dev(rt2x00dev);
	if (retval)
		goto exit_free_device;

	return 0;

exit_free_device:
	ieee80211_free_hw(hw);

exit_put_device:
	usb_put_dev(usb_dev);

	usb_set_intfdata(usb_intf, NULL);

	return retval;
}
EXPORT_SYMBOL_GPL(rt2x00usb_probe);

void rt2x00usb_disconnect(struct usb_interface *usb_intf)
{
	struct ieee80211_hw *hw = usb_get_intfdata(usb_intf);
	struct rt2x00_dev *rt2x00dev = hw->priv;

	/*
	 * Uninitialize and free the rt73usb driver data.
	 */
	rt2x00lib_disable_radio(rt2x00dev);
	rt2x00usb_uninitialize(rt2x00dev);
	rt2x00usb_remove_dev(rt2x00dev);

	/*
	 * Free the 80211 stack data.
	 */
	ieee80211_free_hw(hw);

	/*
	 * Free the USB device data.
	 */
	usb_set_intfdata(usb_intf, NULL);
	usb_put_dev(interface_to_usbdev(usb_intf));
}
EXPORT_SYMBOL_GPL(rt2x00usb_disconnect);

#ifdef CONFIG_PM
int rt2x00usb_suspend(struct usb_interface *usb_intf, pm_message_t state)
{
	struct ieee80211_hw *hw = usb_get_intfdata(usb_intf);
	struct rt2x00_dev *rt2x00dev = hw->priv;
	int retval;

	NOTICE("Going to sleep.\n");

	/*
	 * Disable the radio.
	 */
	rt2x00lib_disable_radio(rt2x00dev);

	/*
	 * Set device mode to sleep for power management.
	 */
	retval = rt2x00dev->ops->lib->set_state(rt2x00dev, STATE_SLEEP);
	if (retval)
		return retval;

	/*
	 * Uninitialize device and hardware.
	 */
	rt2x00usb_uninitialize(rt2x00dev);
	rt2x00usb_remove_dev(rt2x00dev);

	/*
	 * Decrease usbdev refcount.
	 */
	usb_put_dev(interface_to_usbdev(usb_intf));

	return 0;
}
EXPORT_SYMBOL_GPL(rt2x00usb_suspend);

int rt2x00usb_resume(struct usb_interface *usb_intf)
{
	struct ieee80211_hw *hw = usb_get_intfdata(usb_intf);
	struct rt2x00_dev *rt2x00dev = hw->priv;
	int retval;

	NOTICE("Waking up.\n");

	/*
	 * Increase usbdev refcount.
	 */
	usb_get_dev(interface_to_usbdev(usb_intf));

	/*
	 * Initialize hardware.
	 */
	retval = rt2x00usb_probe_dev(rt2x00dev);
	if (retval) {
		ERROR("Failed to allocate device.\n");
		return retval;
	}

	/*
	 * Set device mode to awake for power management.
	 */
	retval = rt2x00dev->ops->lib->set_state(rt2x00dev, STATE_AWAKE);
	if (retval)
		return retval;

	return 0;
}
EXPORT_SYMBOL_GPL(rt2x00usb_resume);
#endif /* CONFIG_PM */

/*
 * rt2x00usb module information.
 */
static char version[] =
	DRV_NAME " - " DRV_VERSION " (" DRV_RELDATE ") by " DRV_PROJECT;

MODULE_AUTHOR(DRV_PROJECT);
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION("rt2x00 library");
MODULE_LICENSE("GPL");

static int __init rt2x00usb_init(void)
{
	printk(KERN_INFO "Loading module: %s.\n", version);
	return 0;
}

static void __exit rt2x00usb_exit(void)
{
	printk(KERN_INFO "Unloading module: %s.\n", version);
}

module_init(rt2x00usb_init);
module_exit(rt2x00usb_exit);
